import asyncio
import codecs
import json
import logging
import os
import socket
import sys
import typing
from typing import Optional
from termcolor import colored

import requests
from pydantic import BaseModel, ValidationError

if typing.TYPE_CHECKING:
    import websockets

log = logging.getLogger(__name__)


class SessionPayload(BaseModel):
    sessionId: str
    sessionSecret: str


class BodyPayload(BaseModel):
    requestId: str
    path: str
    method: str = "GET"
    body: dict | None = None
    headers: dict


class ExposeSessionJson(BaseModel):
    expose_session: str
    url: str


def get_expose_session_path(datadir: str) -> str:
    return os.path.join(datadir, "expose_session.json")


def read_expose_session_json(datadir: str) -> None | ExposeSessionJson:
    session_json = None
    try:
        expose_session_path = get_expose_session_path(datadir)
        log.debug(f"🗂️ Reading expose_session.json path={expose_session_path}")
        with open(expose_session_path, "r") as f:
            session_json = ExposeSessionJson(**json.load(f))
    except FileNotFoundError:
        pass
    except ValidationError as e:
        log.error("Failed to load previous expose session", e)
    except json.JSONDecodeError as e:
        log.error("Failed to decode exopse session json", e)
    return session_json


def write_expose_session_json(
    datadir: str, expose_session_json: ExposeSessionJson
) -> None:
    expose_session_path = get_expose_session_path(datadir)
    log.debug(f"🗂️ Writing expose_session.json path={expose_session_path}")
    with open(expose_session_path, "w") as f:
        json.dump(
            expose_session_json.model_dump(),
            f,
            indent=2,
        )


def get_expose_session(payload: SessionPayload) -> str:
    return f"{payload.sessionId}:{payload.sessionSecret}".encode("ascii").hex()


def get_expose_session_payload(expose_session: str) -> SessionPayload:
    session_id, session_secret = codecs.decode(
        expose_session.encode("ascii"), "hex"
    ).split(b":")
    return SessionPayload(
        sessionId=session_id.decode(), sessionSecret=session_secret.decode()
    )


def forward_request(base_url: str, payload: BodyPayload) -> requests.Response:
    url = base_url.rstrip("/") + "/" + payload.path.lstrip("/")

    if payload.method == "GET":
        return requests.get(url, headers=payload.headers)
    elif payload.method == "POST":
        return requests.post(url, json=payload.body, headers=payload.headers)
    elif payload.method == "PUT":
        return requests.put(url, json=payload.body, headers=payload.headers)
    elif payload.method == "DELETE":
        return requests.delete(url, headers=payload.headers)
    else:
        raise NotImplementedError(f"Method {payload.method} not implemented")


async def handle_ping_pong(
    ws: "websockets.WebSocketClientProtocol",
    pong_queue: asyncio.Queue,
    ping_interval: int,
):
    while True:
        await asyncio.sleep(ping_interval)
        await ws.send("ping")

        try:
            await asyncio.wait_for(pong_queue.get(), timeout=ping_interval)
        except (asyncio.TimeoutError, TimeoutError):
            log.debug("Lost connection to expose server")
            await ws.close()
            break


def handle_session_payload(
    session_payload: SessionPayload, expose_url: str, datadir: str, api_key: str | None
):
    url = f"https://{session_payload.sessionId}.{expose_url}"
    log.info(
        colored("  🌍 Public URL: ", "green", attrs=["bold"])
        + colored(f"{url}", "light_blue")
    )
    if api_key:
        log.info(
            colored("  🔑 API Authorization Bearer key: ", attrs=["bold"])
            + f"{api_key}\n"
        )
    new_expose_session = get_expose_session(session_payload)
    write_expose_session_json(
        datadir=datadir,
        expose_session_json=ExposeSessionJson(
            expose_session=new_expose_session,
            url=url,
        ),
    )


def handle_body_payload(
    payload: BodyPayload,
    base_url: str,
):
    response: requests.Response = forward_request(base_url=base_url, payload=payload)
    return json.dumps(
        {
            "requestId": payload.requestId,
            "response": json.dumps(
                response.json(),
                indent=2,
            ),
            "status": response.status_code,
        }
    )


async def expose_server(
    port: int,
    host: str,
    expose_url: str,
    datadir: str,
    expose_session: str | None = None,
    ping_interval: int = 4,
    api_key: str | None = None,
):
    """
    Exposes the server to the world.
    """
    import websockets

    pong_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

    async def listen_for_requests() -> None:
        session_payload: Optional[SessionPayload] = (
            get_expose_session_payload(expose_session) if expose_session else None
        )

        headers = (
            {
                "x-session-id": session_payload.sessionId,
                "x-session-secret": session_payload.sessionSecret,
            }
            if session_payload
            else {}
        )

        async for ws in websockets.connect(
            f"wss://client.{expose_url}",
            extra_headers=headers,
            logger=log,
            open_timeout=2,
            close_timeout=0,
        ):
            ping_task = asyncio.create_task(
                handle_ping_pong(ws, pong_queue, ping_interval)
            )

            try:
                while True:
                    message = await ws.recv()
                    if message == "pong":
                        await pong_queue.put("pong")
                        continue

                    data = json.loads(message)

                    match data:
                        case {"sessionId": _, "sessionSecret": _}:
                            try:
                                session_payload = SessionPayload(**data)
                                handle_session_payload(
                                    session_payload, expose_url, datadir, api_key
                                )
                            except ValidationError as e:
                                if not session_payload:
                                    log.error("Expose session initialization failed", e)
                                    continue
                                pass
                        case _:
                            try:
                                body_payload = BodyPayload(**data)
                                response = handle_body_payload(
                                    body_payload, f"http://{host}:{port}"
                                )
                                await ws.send(response)
                            except ValidationError as e:
                                log.error("Expose request validation failed", e)
                                continue
            except websockets.exceptions.ConnectionClosedError:
                log.info("Lost connection. Reconnecting to expose server...")
            except socket.gaierror:
                log.info("No internet connection")
                break
            finally:
                ping_task.cancel()
                try:
                    await ping_task
                except asyncio.CancelledError as e:
                    pass
        else:
            log.info("Expose server connection closed")

    task = asyncio.create_task(listen_for_requests())
    await task  # Wait for listen_for_requests to complete


def main(parent_pid, port, verbose, host, expose_url, datadir, expose_session, api_key):
    from robocorp.action_server._robo_utils.process import exit_when_pid_exists

    logging.basicConfig(
        level=logging.DEBUG if verbose.count("v") > 0 else logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
    )

    exit_when_pid_exists(int(parent_pid))
    try:
        asyncio.run(
            expose_server(
                port=int(port),
                host=host,
                expose_url=expose_url,
                datadir=datadir,
                expose_session=expose_session if expose_session != "None" else None,
                api_key=api_key,
            )
        )
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main(*sys.argv[1:])



import re
from datetime import datetime
from logging import Filter, Formatter

from termcolor import colored


class FormatterNoColor(Formatter):
    strip_colors_regex = re.compile(r"(\x1b\[|\x9b)[0-?]*[ -/]*[@-~]")

    def format(self, record):
        message = super().format(record)
        return self.strip_colors_regex.sub("", message)


class FormatterStdout(Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from uvicorn.logging import AccessFormatter

        self._access_formatter = AccessFormatter()

    def format(self, record):
        if record.name.startswith("uvicorn"):
            (
                client_addr,
                method,
                full_path,
                http_version,
                status_code,
            ) = record.args

            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            status_code = self._access_formatter.get_status_code(int(status_code))
            return (
                colored(f"{timestamp}: ", attrs=["dark"])
                + colored(method, attrs=["bold"])
                + f" {full_path} {status_code}"
            )

        return super().format(record)


class UvicornLogFilter(Filter):
    def filter(self, record):
        # Log only Uvicorn Access messages
        return not (record.name.startswith("uvicorn") and len(record.args) != 5)



import datetime
import inspect
import json
import logging
import time
import typing
from typing import Annotated, Any, Dict, Optional

from fastapi import HTTPException, Response
from fastapi.params import Header, Param
from pydantic import BaseModel

if typing.TYPE_CHECKING:
    from ._models import Action, Run

log = logging.getLogger(__name__)


_spec_api_type_to_python_type_str = {
    "string": "str",
    "integer": "int",
    "number": "float",
    "boolean": "bool",
}

_spec_api_type_to_python_type = {
    "string": str,
    "integer": int,
    "number": float,
    "boolean": bool,
}


def _create_run_artifacts_dir(action: "Action", run_id: str) -> str:
    """
    Returns:
        The path, relative to the settings.artifacts_dir which should be used
        to store the output of the given run.
    """
    from ._settings import get_settings

    settings = get_settings()
    artifacts_dir = settings.artifacts_dir
    path = run_id
    target_dir = artifacts_dir / path
    target_dir.mkdir(parents=True, exist_ok=False)
    return path


def _create_run(
    action: "Action", run_id: str, inputs: dict, relative_artifacts_dir: str
) -> "Run":
    from robocorp.action_server._models import RUN_ID_COUNTER, Counter

    from ._database import datetime_to_str
    from ._models import Run, RunStatus, get_db
    from ._runs_state_cache import get_global_runs_state

    db = get_db()
    run_kwargs: Dict[str, Any] = dict(
        id=run_id,
        status=RunStatus.NOT_RUN,
        action_id=action.id,
        start_time=datetime_to_str(datetime.datetime.now(datetime.timezone.utc)),
        run_time=None,
        inputs=json.dumps(inputs),
        result=None,
        error_message=None,
        relative_artifacts_dir=relative_artifacts_dir,
    )
    with db.transaction():
        with db.cursor() as cursor:
            db.execute_update_returning(
                cursor,
                "UPDATE counter SET value=value+1 WHERE id=? RETURNING value",
                [RUN_ID_COUNTER],
            )
            counter_record = cursor.fetchall()
            if not counter_record:
                raise RuntimeError(
                    f"Error. No counter found for run_id. Counters in db: {db.all(Counter)}"
                )
            run_kwargs["numbered_id"] = counter_record[0][0]

        run = Run(**run_kwargs)
        db.insert(run)

    # Ok, transaction finished properly. Let's add it to our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_inserted(run)

    return run


def _update_run(run: "Run", initial_time: float, run_finished: bool, **changes):
    from ._models import get_db
    from ._runs_state_cache import get_global_runs_state

    if run_finished:
        changes["run_time"] = time.monotonic() - initial_time

    db = get_db()
    for k, v in changes.items():
        setattr(run, k, v)
    fields_changed = tuple(changes.keys())
    with db.transaction():
        db.update(run, *fields_changed)

    # Ok, transaction finished properly. Let's update our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_changed(run, changes)


def _set_run_as_finished_ok(run: "Run", result: str, initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, True, result=result, status=RunStatus.PASSED)


def _set_run_as_finished_failed(run: "Run", error_message: str, initial_time: float):
    from ._models import RunStatus

    _update_run(
        run, initial_time, True, status=RunStatus.FAILED, error_message=error_message
    )


def _set_run_as_running(run: "Run", initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, False, status=RunStatus.RUNNING)


def _run_action_in_thread(
    action: "Action",
    response: Response,
    signature: inspect.Signature,
    headers: Dict[str, str],
    *args,
    **kwargs,
):
    """
    This is where the user actually runs something.

    This runs in a thread (so, be careful when talking to the database).

    We have to take care of making a run with the proper environment,
    creating the run, collecting output info, etc.
    """
    from robocorp.action_server._gen_ids import gen_uuid
    from robocorp.action_server._settings import get_settings

    from ._actions_process_pool import get_actions_process_pool
    from ._models import Run, get_db

    settings = get_settings()
    ret_type = kwargs.pop("__ret_type__", "")
    if not ret_type:
        ret_type = "string"

    db = get_db()
    with db.connect():  # Connection is per-thread, so, we need to create a new one.
        inputs = {}
        if args or kwargs:
            bound = signature.bind(*args, **kwargs)

            for k, v in bound.arguments.items():
                inputs[k] = v

        expected_return_type = _spec_api_type_to_python_type.get(ret_type)
        if not expected_return_type:
            raise RuntimeError(
                f"The return type in the spec for the action {action.name} is not valid. Found: {ret_type}."
            )

        actions_process_pool = get_actions_process_pool()
        reuse_process = settings.reuse_processes

        with actions_process_pool.obtain_process_for_action(action) as process_handle:
            run_id = gen_uuid("run")
            response.headers["X-Action-Server-Run-Id"] = run_id
            relative_artifacts_path: str = _create_run_artifacts_dir(action, run_id)
            run: Run = _create_run(action, run_id, inputs, relative_artifacts_path)

            input_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_inputs.json"
            )
            input_json.write_text(json.dumps(inputs))

            robot_artifacts = settings.artifacts_dir / relative_artifacts_path

            result_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_result.json"
            )

            output_file = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_output.txt"
            )

            initial_time = time.monotonic()
            try:
                _set_run_as_running(run, initial_time)
                returncode = process_handle.run_action(
                    action,
                    input_json,
                    robot_artifacts,
                    output_file,
                    result_json,
                    headers,
                    reuse_process,
                )

                if returncode == 0:
                    try:
                        result: str = result_json.read_text("utf-8", "replace")
                    except Exception:
                        raise RuntimeError(
                            "It was not possible to collect the contents of the "
                            "result (json not created)."
                        )
                    try:
                        _set_run_as_finished_ok(run, result, initial_time)
                        ret = json.loads(result)
                    except Exception:
                        raise RuntimeError(
                            "Error loading the contents of {result_json} as json."
                        )
                    else:
                        if not isinstance(ret, expected_return_type):
                            raise RuntimeError(
                                f"Error in action. Expected return type: {ret_type}. Found return type: {type(ret)} (value: {ret})."
                            )
                        return ret

                raise RuntimeError(
                    f"Error: the process did not complete successfully when "
                    f"running. returncode: {returncode}"
                )

            except BaseException as e:
                _set_run_as_finished_failed(run, str(e), initial_time)
                raise HTTPException(status_code=500, detail=str(e))


def _name_as_class_name(name):
    return name.replace("_", " ").title().replace(" ", "")


def generate_func_from_action(action: "Action"):
    """
    This function generates a method from the given action.

        class CalculatorSumInput(BaseModel):
            v1: Annotated[float, Param(description='First number.')]
            v2: Annotated[float, Param(description='Second number.')]

        def calculator_sum_as_params(
            v1: Annotated[float, Param(description='First number.')],
            v2: Annotated[float, Param(description='Second number.')]
        )-> Annotated[float, Param(description='The sum of v1 + v2.')]:

            pass

        def calculator_sum(
            args:CalculatorSumInput,
            response: Response,
            x_action_trace: Optional[str] = Header(None, description='Client application run trace reference', alias='X-action-trace', include_in_schema=False)
        ) -> Annotated[float, Param(description='The sum of v1 + v2.')]:

            headers = {'x_action_trace': x_action_trace}
            return _run_action_in_thread(action, signature, headers, args.v1, args.v2, __ret_type__='number')

        signature = inspect.signature(calculator_sum_as_params)
    """
    input_schema_dict = json.loads(action.input_schema)
    output_schema_dict = json.loads(action.output_schema)
    properties = input_schema_dict.get("properties", {})
    arguments = []
    argument_names = []
    for param_name, param_data in properties.items():
        desc = param_data.get("description", "")
        argument_names.append(f"args.{param_name}")
        param_type = param_data.get("type", "string")

        argument = f"""
    {param_name}: Annotated[{_spec_api_type_to_python_type_str[param_type]}, 
        Param(description={desc!r})]"""
        if "default" in param_data:
            default = param_data.get("default")
            argument = f"{argument}={default!r}"
        arguments.append(argument)

    # Note: Headers are explicitly hidden from spec to make the OpenAPI schema compatible with OpenAI
    headers = {
        "x_action_trace": "Optional[str] = Header(None, description='Client application run trace reference', alias='X-action-trace', include_in_schema=False)"
    }
    headers_as_params = [f"{key}: {value}" for key, value in headers.items()]
    headers_as_values = [f"'{key}': {key}" for key, _ in headers.items()]

    ret_type = output_schema_dict.get("type", "")
    ret = ""
    if ret_type:
        ret_desc = output_schema_dict.get("description")
        ret = f"""-> Annotated[{_spec_api_type_to_python_type_str[ret_type]}, 
        Param(description={ret_desc!r})]"""
    argument_names.append(f"__ret_type__={ret_type!r}")

    code = f"""
class {_name_as_class_name(action.name)}Input(BaseModel):
    pass
{''.join(arguments)}
    
def {action.name}_as_params({', '.join(arguments)}){ret}:
    pass

def {action.name}(args:{_name_as_class_name(action.name)}Input, response: Response, {", ".join(headers_as_params)}){ret}:
    headers = {{{", ".join(headers_as_values)}}}
    return _run_action_in_thread(action, response, signature, headers, {', '.join(argument_names)})

signature = inspect.signature({action.name}_as_params)
"""
    # print(f"Code:\n{code}")
    compiled = compile(code, "<string>", "exec")
    ctx: dict = {
        "Annotated": Annotated,
        "Param": Param,
        "Header": Header,
        "BaseModel": BaseModel,
        "Optional": Optional,
        "action": action,
        "inspect": inspect,
        "Response": Response,
        "_run_action_in_thread": _run_action_in_thread,
    }
    exec(compiled, ctx)
    method = ctx[action.name]
    return method



from pathlib import Path


def generate_api_key():
    import secrets

    return secrets.token_urlsafe(32)


def get_api_key(datadir: Path) -> str:
    api_key_path = datadir / ".api_key"

    try:
        api_key = api_key_path.read_text("utf-8")
    except Exception:
        api_key = ""

    if len(api_key) == 0:
        api_key = generate_api_key()
        api_key_path.write_text(api_key, "utf-8")

    return api_key
